use std::str::FromStr;
use std::{env, sync::Mutex};

use actix_cors::Cors;
use actix_web::middleware::Logger;
use actix_web::{http, web, App, HttpServer};
use cron::Schedule;
use dotenvy::dotenv;
use env_logger::Env;
use middleware::scheduler::du::scan_and_write_disk_usage_to_redis;

mod connections;
mod middleware;
mod model;
mod result;
mod router;
mod schema;

use connections::{CacheInit, Connections, ConnectionsInit, DatabaseInit};
use middleware::scheduler::start_scheduler;

#[macro_use]
extern crate diesel;

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    dotenv().ok();
    let database_url = env::var("DATABASE_URL").expect("DATABASE_URL must be set");
    let cache_url = env::var("CACHE_URL").expect("CACHE_URL must be set");
    let frontend_url = env::var("FRONTEND_URL").expect("FRONTEND_URL must be set");

    let connections = web::Data::new(Mutex::new(
        Connections::new(ConnectionsInit {
            db_init: DatabaseInit { url: database_url },
            cache_init: CacheInit { url: cache_url },
        })
        .unwrap(),
    ));

    // Scheduler example
    let schedule = Schedule::from_str("0 * * * * * *").unwrap();
    start_scheduler(schedule, &connections, |connections| {
        let conns = connections.lock().unwrap();
        let mut cache_conn = conns.cache.get().unwrap();
        let mut db_connection = conns.db.get().unwrap();
        match scan_and_write_disk_usage_to_redis(&mut db_connection, &mut cache_conn) {
            Ok(()) => true,
            Err(error) => {
                eprintln!("Failed to scan disk usage and write to redis cache: {error}");
                false
            }
        }
    });

    env_logger::init_from_env(Env::default().default_filter_or("info"));

    HttpServer::new(move || {
        let cors = Cors::default()
            .allowed_origin(&frontend_url)
            .supports_credentials()
            .allowed_headers(vec![
                http::header::AUTHORIZATION,
                http::header::CONTENT_TYPE,
            ])
            .allowed_methods(vec!["POST", "GET", "PATCH", "OPTIONS", "DELETE"]);
        App::new()
            .wrap(cors)
            .wrap(Logger::default())
            .wrap(Logger::new("%a %{User-Agent}i"))
            .app_data(web::Data::clone(&connections))
            .service(crate::router::users::add_user_and_team)
            .service(crate::router::project_requests::insert)
    })
    .bind(("127.0.0.1", 8080))?
    .run()
    .await?;
    Ok(())
}
