use lettre::{
    message::{header::ContentType, Mailbox},
    transport::smtp::{
        authentication::Credentials,
        client::{Tls, TlsParameters},
        response::Response,
        Error,
    },
    Message, SmtpTransport, Transport,
};
use tera::Tera;

/// Returns the template to send as an email with all the variables replaced
///
/// # Arguments
///
/// *`template_name` - The name of the template used, it should match a file in the file directory tree e.g. "template.html". **<span style="color:red;">All the templates should be located in src/middleware/emails/template</span>**
///
/// *`context` - The context of the template, it should hold all of the key-value pairs to replace in the template
#[allow(dead_code)]
fn get_email_template(template_name: &str, context: tera::Context) -> String {
    let template_dir = "src/middleware/emails/template/**/*";
    let tera = Tera::new(template_dir).unwrap();
    tera.render(template_name, &context).unwrap()
}

/// Send an email using all the informations in input
///
/// # Arguments
///
/// *`smtp_server_url` - The url of the smtp server used to relay the email
///
/// *`smtp_server_port` - The port of the smtp server used to relay the email
///
/// *`credentials` - The username-password pair of the user used to log into the smtp server to send the email
///
/// *`from` - The sender of the email
///
/// *`to` - The recipient of the email
///
/// *`subject` - The subject of the email
///
/// *`content` - The content of the email, usually the output of the [get_email_template]
#[allow(dead_code)]
fn send_email(
    smtp_server_url: &str,
    smtp_server_port: u16,
    credentials: Credentials,
    from: Mailbox,
    to: Mailbox,
    subject: &str,
    content: &str,
) -> Result<Response, Error> {
    let email = Message::builder()
        .from(from)
        .to(to)
        .subject(subject)
        .header(ContentType::TEXT_HTML)
        .body(String::from(content))
        .unwrap();

    let tls = TlsParameters::builder(smtp_server_url.to_string())
        .dangerous_accept_invalid_certs(true)
        .build()
        .unwrap();
    let mailer = SmtpTransport::starttls_relay(smtp_server_url)
        .unwrap()
        .credentials(credentials)
        .port(smtp_server_port)
        .tls(Tls::Required(tls))
        .build();
    mailer.send(&email)
}

#[cfg(test)]
mod tests {
    use dotenvy::dotenv;
    use lettre::{message::Mailbox, transport::smtp::authentication::Credentials};
    use std::env;
    use tera::Context;

    use super::{get_email_template, send_email};

    #[test]
    fn test() {
        dotenv().ok();
        let smtp_server_url = env::var("SMTP_SERVER_URL").expect("SMTP_SERVER_URL must be set");
        let smtp_server_port: u16 = env::var("SMTP_SERVER_PORT")
            .expect("SMTP_SERVER_PORT must be set")
            .parse()
            .unwrap();
        let smtp_mailer_email =
            env::var("SMTP_MAILER_EMAIL").expect("SMTP_MAILER_EMAIL must be set");
        let smtp_mailer_password =
            env::var("SMTP_MAILER_PASSWORD").expect("SMTP_MAILER_PASSWORD must be set");
        let smtp_test_recipient =
            env::var("TEST_SMTP_EMAIL_RECIPIENT").expect("SMTP_TEST_EMAIL_RECIPIENT must be set");

        let credentials = Credentials::new(smtp_mailer_email.clone(), smtp_mailer_password);
        let from: Mailbox = smtp_mailer_email.parse().unwrap();
        let to: Mailbox = smtp_test_recipient.parse().unwrap();
        let template_name = "quota_reached.html";
        let mut context = Context::new();
        context.insert("username", "Another username");
        let email_content = get_email_template(template_name, context);
        send_email(
            &smtp_server_url,
            smtp_server_port,
            credentials,
            from,
            to,
            "Test mail",
            &email_content,
        )
        .unwrap();
    }
}
