use camino::Utf8Path;
use std::io::BufRead;
use std::process::{Command, Stdio};
use which::which;

use crate::result::Error;

/// Returns whether du is installed on the system or not
fn is_du_installed() -> bool {
    match which("du") {
        Ok(_pathbuff) => true,
        Err(_error) => false,
    }
}

/// Uses du to calculate the total size of a directory and its content on disk
///
/// # Arguments
///
/// *`directory` - The path of the folder we want to know the size of
///
/// * Returns a Result containing either an error message or a u64 representing the size of the directory in kilobytes.
///
pub fn du(directory: &Utf8Path) -> Result<u64, Error> {
    if !is_du_installed() {
        return Err(Error::PackageNotInstalled(String::from("du")));
    }
    let mut progress = Command::new("du")
        .args(["-sx", directory.as_str()])
        .stdout(Stdio::piped())
        .spawn()
        .unwrap();

    progress.wait().unwrap();

    let mut res = String::new();
    if let Some(progress) = progress.stdout.take() {
        let reader = std::io::BufReader::new(progress);
        for line in reader.lines().flatten() {
            res.push_str(&line);
        }
    }
    Ok(res.split('\t').collect::<Vec<&str>>()[0]
        .to_string()
        .parse::<u64>()
        .unwrap())
}

#[cfg(test)]
mod tests {
    use camino::Utf8PathBuf;
    use tempfile::tempdir;

    use super::du;

    #[test]
    fn test_du() {
        let temp_dir = tempdir().unwrap();
        let path = Utf8PathBuf::from_path_buf((*temp_dir.as_ref()).to_path_buf()).unwrap();
        let result = du(&path).unwrap();
        assert_eq!(result, 4);
        let file = &temp_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world touch").unwrap();
        let result_with_file = du(&path).unwrap();
        assert_eq!(result_with_file, 8);

        drop(temp_dir);
    }
}
