use std::collections::HashMap;

use diesel::PgConnection;
use redis::{Commands, RedisResult};
use uuid::Uuid;

use crate::{
    middleware::filesystem::du::du,
    model::operations::{
        datasets::find_by_project,
        projects::{find_all, find_by_id},
    },
    result::Error,
};

/// Returns a hashmap of all the datasets with their corresponding size, the key being the id of the dataset and the value the size of the folder in kilobytes
/// # Arguments
///
/// * `project_id` - The id of the project we want to scan
///
/// * `connection` - The connection to the database where the projects and datasets are stored
///
pub fn scan_project(
    connection: &mut PgConnection,
    project_id: Uuid,
) -> Result<HashMap<Uuid, u64>, Error> {
    match find_by_id(connection, project_id) {
        Ok(option_project) => {
            let project = option_project.unwrap();
            match find_by_project(connection, project_id) {
                Ok(datasets) => {
                    let mut result = HashMap::new();
                    for dataset in datasets {
                        let dataset_absolute_path = project.path.join(dataset.path);
                        let dataset_du = du(&dataset_absolute_path);
                        match dataset_du {
                            Ok(dataset_size) => {
                                result.insert(dataset.id, dataset_size);
                            }
                            Err(error) => return Err(error),
                        }
                    }
                    Ok(result)
                }
                Err(error) => Err(Error::Model(error)),
            }
        }
        Err(error) => Err(Error::Model(error)),
    }
}

/// Returns a hashmap of all the projects and the belonging datasets with their corresponding size, the key is the id of the project, see [`scan_project`] for the content of the value
/// # Arguments
///
/// * `connection` - The connection to the database where the projects and datasets are stored
///
pub fn scan_projects(
    connection: &mut PgConnection,
) -> Result<HashMap<Uuid, HashMap<Uuid, u64>>, Error> {
    let mut result = HashMap::new();
    match find_all(connection) {
        Ok(projects) => {
            for project in projects {
                let datasets_hashmap = scan_project(connection, project.id)?;
                result.insert(project.id, datasets_hashmap);
            }
        }
        Err(error) => return Err(Error::Model(error)),
    }
    Ok(result)
}

/// Scans all the datasets in all the projects and writes the disk usage of each of the datasets in a redis cache
///
/// # Arguments
///
/// * `db_connection` - The connection to the database where the projects and datasets are stored
///
/// * `redis_connection` - The connection to the cache where the disk usage informations are stored
///
pub fn scan_and_write_disk_usage_to_redis(
    db_connection: &mut PgConnection,
    redis_connection: &mut redis::Connection,
) -> Result<(), Error> {
    let projects_disk_usage_hashmap = scan_projects(db_connection)?;
    for (project_id, datasets_disk_usage_hashmap) in projects_disk_usage_hashmap {
        let key = format!("du_{project_id}");
        for (dataset_id, dataset_disk_usage_hashmap) in datasets_disk_usage_hashmap {
            let result: RedisResult<()> = redis_connection.hset(
                key.clone(),
                dataset_id.to_string(),
                dataset_disk_usage_hashmap.to_string(),
            );
            if let Err(error) = result {
                return Err(Error::Cache(error));
            }
        }
    }

    Ok(())
}

#[cfg(test)]
mod tests {

    use std::env::{self, temp_dir};

    use camino::Utf8PathBuf;
    use diesel::{PgConnection, QueryResult};
    use dotenvy::dotenv;
    use redis::Commands;
    use tempfile::{tempdir, tempdir_in};
    use uuid::Uuid;

    use crate::model::{
        operations::tests::establish_connection,
        structures::{datasets::Dataset, projects::Project},
    };

    use super::{scan_and_write_disk_usage_to_redis, scan_project, scan_projects};

    fn create_dataset(
        connection: &mut PgConnection,
        path: Utf8PathBuf,
        project_id: Option<Uuid>,
    ) -> QueryResult<Dataset> {
        let dataset = Dataset {
            id: Uuid::new_v4(),
            path,
            project_id,
        };
        crate::model::operations::datasets::insert_with_id(connection, dataset)
    }

    fn create_project(
        connection: &mut PgConnection,
        name: String,
        path: Utf8PathBuf,
        user_id: Option<i32>,
    ) -> QueryResult<Project> {
        let project = Project {
            id: Uuid::new_v4(),
            name,
            path,
            user_id,
        };
        crate::model::operations::projects::insert_with_id(connection, project)
    }

    #[test]
    fn test_scan_project() {
        dotenv().ok();
        let mut connection = establish_connection();
        let temp_dir = tempdir().unwrap();
        let project_path = Utf8PathBuf::from_path_buf(temp_dir.path().to_owned()).unwrap();
        let dataset_dir = tempdir_in(&temp_dir).unwrap();
        let dataset_path = Utf8PathBuf::from_path_buf(dataset_dir.path().to_owned()).unwrap();
        let another_dataset_dir = tempdir_in(&temp_dir).unwrap();
        let another_dataset_path =
            Utf8PathBuf::from_path_buf(another_dataset_dir.path().to_owned()).unwrap();
        let project = create_project(
            &mut connection,
            String::from("Project 1"),
            project_path,
            None,
        )
        .unwrap();
        create_dataset(&mut connection, dataset_path, Some(project.id)).unwrap();
        create_dataset(&mut connection, another_dataset_path, Some(project.id)).unwrap();
        let file = &dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();

        let datasets_hashmap = scan_project(&mut connection, project.id).unwrap();
        assert_eq!(datasets_hashmap.len(), 2);
        // Hashmap contains the empty folder:
        assert!(datasets_hashmap.values().any(|&value| value == 4));
        // Hashmap contains the folder with a file:
        assert!(datasets_hashmap.values().any(|&value| value == 8));
        // We drop the directories manually otherwise they are deleted before the scan
        drop(dataset_dir);
        drop(another_dataset_dir);
        drop(temp_dir);
    }

    #[test]
    fn test_scan_projects() {
        dotenv().ok();
        let mut connection = establish_connection();

        //First project
        let alpha_temp_dir: tempfile::TempDir = tempdir().unwrap();
        let alpha_project_path =
            Utf8PathBuf::from_path_buf(alpha_temp_dir.path().to_owned()).unwrap();
        let alpha_dataset_dir = tempdir_in(&alpha_temp_dir).unwrap();
        let alpha_dataset_path =
            Utf8PathBuf::from_path_buf(alpha_dataset_dir.path().to_owned()).unwrap();
        let alpha_another_dataset_dir = tempdir_in(&alpha_temp_dir).unwrap();
        let alpha_another_dataset_path =
            Utf8PathBuf::from_path_buf(alpha_another_dataset_dir.path().to_owned()).unwrap();
        let alpha_project = create_project(
            &mut connection,
            String::from("Project alpha"),
            alpha_project_path,
            None,
        )
        .unwrap();
        let alpha_project_first_dataset =
            create_dataset(&mut connection, alpha_dataset_path, Some(alpha_project.id)).unwrap();
        let alpha_project_second_dataset = create_dataset(
            &mut connection,
            alpha_another_dataset_path,
            Some(alpha_project.id),
        )
        .unwrap();
        let file = &alpha_dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();

        // Second project
        let beta_temp_dir: tempfile::TempDir = tempdir().unwrap();
        let beta_project_path =
            Utf8PathBuf::from_path_buf(beta_temp_dir.path().to_owned()).unwrap();
        let beta_dataset_dir = tempdir_in(&beta_temp_dir).unwrap();
        let beta_dataset_path =
            Utf8PathBuf::from_path_buf(beta_dataset_dir.path().to_owned()).unwrap();
        let beta_second_dataset_dir = tempdir_in(&beta_temp_dir).unwrap();
        let beta_second_dataset_path =
            Utf8PathBuf::from_path_buf(beta_second_dataset_dir.path().to_owned()).unwrap();
        let beta_third_dataset_dir = tempdir_in(&beta_temp_dir).unwrap();
        let beta_third_dataset_path =
            Utf8PathBuf::from_path_buf(beta_third_dataset_dir.path().to_owned()).unwrap();
        let beta_project = create_project(
            &mut connection,
            String::from("Project beta"),
            beta_project_path,
            None,
        )
        .unwrap();
        let beta_project_first_dataset =
            create_dataset(&mut connection, beta_dataset_path, Some(beta_project.id)).unwrap();
        let beta_project_second_dataset = create_dataset(
            &mut connection,
            beta_second_dataset_path,
            Some(beta_project.id),
        )
        .unwrap();
        let beta_project_third_dataset = create_dataset(
            &mut connection,
            beta_third_dataset_path,
            Some(beta_project.id),
        )
        .unwrap();
        let file = &beta_dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();
        let file = &beta_third_dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();
        let file = &beta_third_dataset_dir.path().join("other_data.txt");
        std::fs::write(file, "Hello world").unwrap();

        let projects_hashmap = scan_projects(&mut connection).unwrap();
        assert_eq!(projects_hashmap.len(), 2);
        assert_eq!(
            *projects_hashmap
                .get(&alpha_project.id)
                .unwrap()
                .get(&alpha_project_first_dataset.id)
                .unwrap(),
            8
        );
        assert_eq!(
            *projects_hashmap
                .get(&alpha_project.id)
                .unwrap()
                .get(&alpha_project_second_dataset.id)
                .unwrap(),
            4
        );

        assert_eq!(
            *projects_hashmap
                .get(&beta_project.id)
                .unwrap()
                .get(&beta_project_first_dataset.id)
                .unwrap(),
            8
        );

        assert_eq!(
            *projects_hashmap
                .get(&beta_project.id)
                .unwrap()
                .get(&beta_project_second_dataset.id)
                .unwrap(),
            4
        );
        assert_eq!(
            *projects_hashmap
                .get(&beta_project.id)
                .unwrap()
                .get(&beta_project_third_dataset.id)
                .unwrap(),
            12
        );
        drop(alpha_dataset_dir);
        drop(beta_dataset_dir);
        drop(beta_second_dataset_dir);
        drop(beta_third_dataset_dir);
        drop(temp_dir());
    }

    #[test]
    fn test_scan_and_write_disk_usage_to_redis() {
        dotenv().ok();
        let mut db_connection = establish_connection();
        let test_cache_url = env::var("TEST_CACHE_URL").expect("TEST_CACHE_URL must be set");
        let client = redis::Client::open(test_cache_url).unwrap();
        let mut redis_connection = client.get_connection().unwrap();
        let _: () = redis::cmd("FLUSHDB").query(&mut redis_connection).unwrap();
        //First project
        let alpha_temp_dir: tempfile::TempDir = tempdir().unwrap();
        let alpha_project_path =
            Utf8PathBuf::from_path_buf(alpha_temp_dir.path().to_owned()).unwrap();
        let alpha_dataset_dir = tempdir_in(&alpha_temp_dir).unwrap();
        let alpha_dataset_path =
            Utf8PathBuf::from_path_buf(alpha_dataset_dir.path().to_owned()).unwrap();
        let alpha_another_dataset_dir = tempdir_in(&alpha_temp_dir).unwrap();
        let alpha_another_dataset_path =
            Utf8PathBuf::from_path_buf(alpha_another_dataset_dir.path().to_owned()).unwrap();
        let alpha_project = create_project(
            &mut db_connection,
            String::from("Project alpha"),
            alpha_project_path,
            None,
        )
        .unwrap();
        let alpha_project_first_dataset = create_dataset(
            &mut db_connection,
            alpha_dataset_path,
            Some(alpha_project.id),
        )
        .unwrap();
        let alpha_project_second_dataset = create_dataset(
            &mut db_connection,
            alpha_another_dataset_path,
            Some(alpha_project.id),
        )
        .unwrap();
        let file = &alpha_dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();

        // Second project
        let beta_temp_dir: tempfile::TempDir = tempdir().unwrap();
        let beta_project_path =
            Utf8PathBuf::from_path_buf(beta_temp_dir.path().to_owned()).unwrap();
        let beta_dataset_dir = tempdir_in(&beta_temp_dir).unwrap();
        let beta_dataset_path =
            Utf8PathBuf::from_path_buf(beta_dataset_dir.path().to_owned()).unwrap();
        let beta_second_dataset_dir = tempdir_in(&beta_temp_dir).unwrap();
        let beta_second_dataset_path =
            Utf8PathBuf::from_path_buf(beta_second_dataset_dir.path().to_owned()).unwrap();
        let beta_third_dataset_dir = tempdir_in(&beta_temp_dir).unwrap();
        let beta_third_dataset_path =
            Utf8PathBuf::from_path_buf(beta_third_dataset_dir.path().to_owned()).unwrap();
        let beta_project = create_project(
            &mut db_connection,
            String::from("Project beta"),
            beta_project_path,
            None,
        )
        .unwrap();
        let beta_project_first_dataset =
            create_dataset(&mut db_connection, beta_dataset_path, Some(beta_project.id)).unwrap();
        let beta_project_second_dataset = create_dataset(
            &mut db_connection,
            beta_second_dataset_path,
            Some(beta_project.id),
        )
        .unwrap();
        let beta_project_third_dataset = create_dataset(
            &mut db_connection,
            beta_third_dataset_path,
            Some(beta_project.id),
        )
        .unwrap();
        let file = &beta_dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();
        let file = &beta_third_dataset_dir.path().join("some_data.txt");
        std::fs::write(file, "Hello world").unwrap();
        let file = &beta_third_dataset_dir.path().join("other_data.txt");
        std::fs::write(file, "Hello world").unwrap();

        scan_and_write_disk_usage_to_redis(&mut db_connection, &mut redis_connection).unwrap();

        let alpha_project_dataset_usage: Result<u64, redis::RedisError> = redis_connection.hget(
            format!("du_{}", alpha_project.id),
            alpha_project_first_dataset.id.to_string(),
        );

        let alpha_project_second_dataset_usage: Result<u64, redis::RedisError> = redis_connection
            .hget(
                format!("du_{}", alpha_project.id),
                alpha_project_second_dataset.id.to_string(),
            );

        let beta_project_first_dataset_usage: Result<u64, redis::RedisError> = redis_connection
            .hget(
                format!("du_{}", beta_project.id),
                beta_project_first_dataset.id.to_string(),
            );
        let beta_project_second_dataset_usage: Result<u64, redis::RedisError> = redis_connection
            .hget(
                format!("du_{}", beta_project.id),
                beta_project_second_dataset.id.to_string(),
            );
        let beta_project_third_dataset_usage: Result<u64, redis::RedisError> = redis_connection
            .hget(
                format!("du_{}", beta_project.id),
                beta_project_third_dataset.id.to_string(),
            );

        assert_eq!(alpha_project_dataset_usage.unwrap(), 8);
        assert_eq!(alpha_project_second_dataset_usage.unwrap(), 4);
        assert_eq!(beta_project_first_dataset_usage.unwrap(), 8);
        assert_eq!(beta_project_second_dataset_usage.unwrap(), 4);
        assert_eq!(beta_project_third_dataset_usage.unwrap(), 12);
    }
}
