pub mod du;
use std::thread::{self, JoinHandle};

use actix_web::web;
use chrono::Local;
use cron::Schedule;

use crate::connections::ConnectionsHandle;

/// Used to execute scheduled tasks
///
/// # Arguments
///
/// *`schedule`: cron-like schedule on which to execute the user task
///
/// *`connections`: ConnectionsHandle to use client connections (like database, cache server, etc) during user tasks. This structure is cloned for you so that you don't have to deal with references in the user task.
///
/// *`user_function`: A closure that takes the ConnectionsHandle as a parameter. The user function has to return a boolean: true to continue running, false to stop itself.
///
pub fn start_scheduler(
    schedule: Schedule,
    connections: &ConnectionsHandle,
    user_function: fn(ConnectionsHandle) -> bool,
) -> JoinHandle<()> {
    let connections = web::Data::clone(connections);
    thread::spawn(move || loop {
        let now = chrono::offset::Local::now();
        let next = schedule.upcoming(Local).take(1);
        let mut wait_time = chrono::Duration::hours(1);
        for n in next {
            wait_time = n - now;
        }
        thread::sleep(wait_time.to_std().unwrap());
        let connections = web::Data::clone(&connections);
        if !user_function(connections) {
            break;
        }
    })
}

#[cfg(test)]
mod tests {
    use std::{env, str::FromStr, sync::Mutex};

    use actix_web::web;
    use cron::Schedule;

    use crate::connections::{CacheInit, Connections, ConnectionsInit, DatabaseInit};

    use super::start_scheduler;

    #[test]
    fn test_simple_one_time_task() {
        let database_url = env::var("DATABASE_URL").expect("DATABASE_URL must be set");
        let cache_url = env::var("TEST_CACHE_URL").expect("TEST_CACHE_URL must be set");

        let connections = web::Data::new(Mutex::new(
            Connections::new(ConnectionsInit {
                db_init: DatabaseInit { url: database_url },
                cache_init: CacheInit { url: cache_url },
            })
            .unwrap(),
        ));

        let schedule = Schedule::from_str("* * * * * * *").unwrap();
        let handle = start_scheduler(schedule, &connections, |_c| {
            false
        });
        assert!(handle.join().is_ok());
    }
}
