use crate::model::structures::datasets::*;
use crate::schema::datasets::dsl::datasets;
use crate::schema::datasets::id;
use diesel::prelude::*;
use uuid::Uuid;

pub fn insert_with_id(
    conn: &mut PgConnection,
    new_dataset: Dataset,
) -> Result<Dataset, diesel::result::Error> {
    use crate::schema::datasets;
    let table_to_update = diesel::insert_into(datasets::table);
    table_to_update
        .values(new_dataset)
        .returning(Dataset::as_returning())
        .get_result(conn)
}

pub fn find_by_id(conn: &mut PgConnection, dataset_id: Uuid) -> QueryResult<Option<Dataset>> {
    datasets
        .filter(id.eq(dataset_id))
        .first::<Dataset>(conn)
        .optional()
}

pub fn find_by_project(conn: &mut PgConnection, project_id: Uuid) -> QueryResult<Vec<Dataset>> {
    let project = match crate::model::operations::projects::find_by_id(conn, project_id) {
        Ok(found_project) => match found_project {
            Some(val) => val,
            None => return Err(diesel::result::Error::NotFound),
        },
        Err(err) => return Err(err),
    };
    let related_datasets = match <Dataset as BelongingToDsl<
        &crate::model::structures::projects::Project,
    >>::belonging_to(&project)
    .load::<Dataset>(conn)
    {
        Ok(result) => result,
        Err(err) => return Err(err),
    };
    Ok(related_datasets)
}

pub fn update(
    conn: &mut PgConnection,
    dataset_id: uuid::Uuid,
    dataset: UpdateDataset,
) -> Result<usize, diesel::result::Error> {
    diesel::update(datasets.filter(id.eq(dataset_id)))
        .set(dataset)
        .execute(conn)
}

pub fn delete(conn: &mut PgConnection, dataset_id: Uuid) -> Result<usize, diesel::result::Error> {
    diesel::delete(datasets.filter(id.eq(dataset_id))).execute(conn)
}

#[cfg(test)]
mod tests {
    use crate::model::{
        operations::{
            datasets::{delete, find_by_id, insert_with_id, update},
            tests::establish_connection,
        },
        structures::datasets::{Dataset, UpdateDataset},
    };
    use camino::Utf8PathBuf;
    use uuid::Uuid;

    #[test]
    fn test_model_insert_with_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_dataset = Dataset {
            id: Uuid::new_v4(),
            path: Utf8PathBuf::from("/tmp/test"),
            project_id: None,
        };
        let inserted_user = insert_with_id(&mut connection, new_dataset.clone()).unwrap();
        assert_eq!(inserted_user.id, new_dataset.id);
        assert_eq!(inserted_user.path, new_dataset.path);
        assert_eq!(inserted_user.project_id, new_dataset.project_id);
    }

    #[test]
    fn test_model_find_by_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_dataset = Dataset {
            id: Uuid::new_v4(),
            path: Utf8PathBuf::from("/tmp/test"),
            project_id: None,
        };
        insert_with_id(&mut connection, new_dataset.clone()).unwrap();
        let found_dataset = find_by_id(&mut connection, new_dataset.id)
            .unwrap()
            .unwrap();
        assert_eq!(found_dataset.id, new_dataset.id);
        assert_eq!(found_dataset.path, new_dataset.path);
        assert_eq!(found_dataset.project_id, new_dataset.project_id);
    }

    #[test]
    fn test_model_update() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_dataset = Dataset {
            id: Uuid::new_v4(),
            path: Utf8PathBuf::from("/tmp/test"),
            project_id: None,
        };
        let updated_dataset = UpdateDataset {
            path: Some(Utf8PathBuf::from("/tmp/test/updated")),
            project_id: None,
        };
        insert_with_id(&mut connection, new_dataset.clone()).unwrap();
        let update_count =
            update(&mut connection, new_dataset.id, updated_dataset.clone()).unwrap();
        let found_dataset = find_by_id(&mut connection, new_dataset.id)
            .unwrap()
            .unwrap();
        assert_eq!(update_count, 1);
        assert_eq!(found_dataset.id, new_dataset.id);
        assert_eq!(found_dataset.path, updated_dataset.path.unwrap());
        assert_eq!(found_dataset.project_id, new_dataset.project_id);
    }

    #[test]
    fn test_model_delete() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_dataset = Dataset {
            id: Uuid::new_v4(),
            path: Utf8PathBuf::from("/tmp/test"),
            project_id: None,
        };
        insert_with_id(&mut connection, new_dataset.clone()).unwrap();
        let dataset_before_delete = find_by_id(&mut connection, new_dataset.id).unwrap();
        assert!(dataset_before_delete.is_some());
        let delete_count = delete(&mut connection, new_dataset.id).unwrap();
        assert_eq!(delete_count, 1);
        let dataset_after_delete = find_by_id(&mut connection, new_dataset.id).unwrap();
        assert!(dataset_after_delete.is_none());
    }
}
