use crate::model::structures::extension_dates::{ExtensionDate, UpdateExtensionDate};
use crate::schema::extension_dates::dsl::extension_dates;
use crate::schema::extension_dates::id;
use diesel::prelude::*;
use uuid::Uuid;

pub fn insert_with_id(
    conn: &mut PgConnection,
    new_extension_date: ExtensionDate,
) -> Result<ExtensionDate, diesel::result::Error> {
    use crate::schema::extension_dates;
    let table_to_update = diesel::insert_into(extension_dates::table);
    table_to_update
        .values(&new_extension_date)
        .returning(ExtensionDate::as_returning())
        .get_result(conn)
}

pub fn find_by_id(
    conn: &mut PgConnection,
    extension_date_id: Uuid,
) -> QueryResult<Option<ExtensionDate>> {
    extension_dates
        .filter(id.eq(extension_date_id))
        .first::<ExtensionDate>(conn)
        .optional()
}

pub fn find_by_project(
    conn: &mut PgConnection,
    project_id: Uuid,
) -> QueryResult<Vec<ExtensionDate>> {
    let project = match crate::model::operations::projects::find_by_id(conn, project_id) {
        Ok(found_project) => match found_project {
            Some(val) => val,
            None => return Err(diesel::result::Error::NotFound),
        },
        Err(err) => return Err(err),
    };
    let related_extension_dates = match <ExtensionDate as BelongingToDsl<
        &crate::model::structures::projects::Project,
    >>::belonging_to(&project)
    .load::<ExtensionDate>(conn)
    {
        Ok(result) => result,
        Err(err) => return Err(err),
    };
    Ok(related_extension_dates)
}

pub fn update(
    conn: &mut PgConnection,
    extension_date_id: uuid::Uuid,
    extension_date: UpdateExtensionDate,
) -> Result<usize, diesel::result::Error> {
    diesel::update(extension_dates.filter(id.eq(extension_date_id)))
        .set(extension_date)
        .execute(conn)
}

pub fn delete(
    conn: &mut PgConnection,
    extension_date_id: Uuid,
) -> Result<usize, diesel::result::Error> {
    diesel::delete(extension_dates.filter(id.eq(extension_date_id))).execute(conn)
}

#[cfg(test)]
mod tests {
    use camino::Utf8PathBuf;
    use chrono::{NaiveDate, NaiveDateTime, NaiveTime};
    use tempfile::tempdir;
    use uuid::Uuid;

    use crate::model::{
        operations::{
            extension_dates::{delete, find_by_id, find_by_project, insert_with_id, update},
            tests::establish_connection,
        },
        structures::{
            extension_dates::{ExtensionDate, UpdateExtensionDate},
            projects::Project,
        },
    };

    #[test]
    fn test_model_insert_with_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let new_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(date, time),
            project_id: None,
        };
        let inserted_extension_date =
            insert_with_id(&mut connection, new_extension_date.clone()).unwrap();
        assert_eq!(new_extension_date, inserted_extension_date);
    }

    #[test]
    fn test_model_find_by_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let new_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(date, time),
            project_id: None,
        };
        let inserted_extension_date = insert_with_id(&mut connection, new_extension_date).unwrap();
        let extension_date_found = find_by_id(&mut connection, inserted_extension_date.id)
            .unwrap()
            .unwrap();
        assert_eq!(extension_date_found, inserted_extension_date);
    }

    #[test]
    fn test_find_by_project() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let temp_dir = tempdir().unwrap();
        let path = Utf8PathBuf::from_path_buf(temp_dir.path().to_owned()).unwrap();
        let project = Project {
            id: Uuid::new_v4(),
            name: String::from("New test project"),
            path,
            user_id: None,
        };

        let first_date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let first_time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let second_date = NaiveDate::from_ymd_opt(2021, 2, 2).unwrap();
        let second_time = NaiveTime::from_hms_milli_opt(1, 1, 1, 1).unwrap();
        let third_date = NaiveDate::from_ymd_opt(2022, 3, 3).unwrap();
        let third_time = NaiveTime::from_hms_milli_opt(2, 2, 2, 2).unwrap();

        // The first two extension dates are part of the project and the third one is not
        let first_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(first_date, first_time),
            project_id: Some(project.id),
        };
        let second_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(second_date, second_time),
            project_id: Some(project.id),
        };
        let third_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(third_date, third_time),
            project_id: None,
        };
        crate::model::operations::projects::insert_with_id(&mut connection, project.clone())
            .unwrap();
        insert_with_id(&mut connection, first_extension_date.clone()).unwrap();
        insert_with_id(&mut connection, second_extension_date.clone()).unwrap();
        insert_with_id(&mut connection, third_extension_date).unwrap();
        let found_extension_dates = find_by_project(&mut connection, project.id).unwrap();
        assert_eq!(&found_extension_dates.len(), &2);
        assert!(found_extension_dates
            .iter()
            .any(|extension_date| extension_date.id == first_extension_date.id));
        assert!(found_extension_dates
            .iter()
            .any(|extension_date| extension_date.id == second_extension_date.id));
    }

    #[test]
    fn test_find_by_project_invalid_project_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let extension_dates = find_by_project(&mut connection, Uuid::new_v4());
        assert!(extension_dates.is_err());
        assert_eq!(extension_dates, Err(diesel::result::Error::NotFound))
    }

    #[test]
    fn test_model_update() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let updated_date = NaiveDate::from_ymd_opt(2021, 2, 2).unwrap();
        let updated_time = NaiveTime::from_hms_milli_opt(1, 1, 1, 1).unwrap();
        let new_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(date, time),
            project_id: None,
        };
        let updated_extension_date = UpdateExtensionDate {
            extension_end_date: Some(NaiveDateTime::new(updated_date, updated_time)),
            project_id: None,
        };
        let inserted_extension_date =
            insert_with_id(&mut connection, new_extension_date.clone()).unwrap();

        let updated_extension_date_count = update(
            &mut connection,
            inserted_extension_date.id,
            updated_extension_date.clone(),
        )
        .unwrap();
        assert_eq!(updated_extension_date_count, 1);

        let saved_extension_date = find_by_id(&mut connection, inserted_extension_date.id)
            .unwrap()
            .unwrap();
        assert_eq!(saved_extension_date.id, new_extension_date.id);
        assert_eq!(
            saved_extension_date.extension_end_date,
            updated_extension_date.extension_end_date.unwrap()
        );
        assert_eq!(
            saved_extension_date.project_id,
            updated_extension_date.project_id
        );
    }

    #[test]
    fn test_model_delete() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let new_extension_date = ExtensionDate {
            id: Uuid::new_v4(),
            extension_end_date: NaiveDateTime::new(date, time),
            project_id: None,
        };
        let inserted_extension_date = insert_with_id(&mut connection, new_extension_date).unwrap();
        find_by_id(&mut connection, inserted_extension_date.id)
            .unwrap()
            .unwrap();

        let rows_affected = delete(&mut connection, inserted_extension_date.id).unwrap();
        assert_eq!(rows_affected, 1);
        let extension_date_found = find_by_id(&mut connection, inserted_extension_date.id).unwrap();
        assert_eq!(extension_date_found, None);
    }
}
