use diesel::prelude::*;
use uuid::Uuid;

use crate::model::structures::project_requests::NewProjectRequest;
use crate::schema::project_requests::dsl::project_requests;
use crate::{
    model::structures::project_requests::{ProjectRequest, UpdateProjectRequest},
    schema::project_requests::id,
};

pub fn insert(
    conn: &mut PgConnection,
    new_project_request: NewProjectRequest,
) -> Result<ProjectRequest, diesel::result::Error> {
    use crate::schema::project_requests;
    let table_to_update = diesel::insert_into(project_requests::table);
    table_to_update
        .values(&new_project_request)
        .get_result(conn)
}

pub fn insert_with_id(
    conn: &mut PgConnection,
    new_project_request: ProjectRequest,
) -> Result<ProjectRequest, diesel::result::Error> {
    use crate::schema::project_requests;
    let table_to_update = diesel::insert_into(project_requests::table);
    table_to_update
        .values(&new_project_request)
        .returning(ProjectRequest::as_returning())
        .get_result(conn)
}

pub fn find_by_id(
    conn: &mut PgConnection,
    project_request_id: Uuid,
) -> QueryResult<Option<ProjectRequest>> {
    project_requests
        .filter(id.eq(project_request_id))
        .first::<ProjectRequest>(conn)
        .optional()
}

pub fn find_by_project(
    conn: &mut PgConnection,
    project_id: Uuid,
) -> QueryResult<Vec<ProjectRequest>> {
    let project = match crate::model::operations::projects::find_by_id(conn, project_id) {
        Ok(found_project) => match found_project {
            Some(val) => val,
            None => return Err(diesel::result::Error::NotFound),
        },
        Err(err) => return Err(err),
    };
    let related_requests = match <ProjectRequest as BelongingToDsl<
        &crate::model::structures::projects::Project,
    >>::belonging_to(&project)
    .load::<ProjectRequest>(conn)
    {
        Ok(result) => result,
        Err(err) => return Err(err),
    };
    Ok(related_requests)
}

pub fn update(
    conn: &mut PgConnection,
    project_request_id: uuid::Uuid,
    project_request: UpdateProjectRequest,
) -> Result<usize, diesel::result::Error> {
    diesel::update(project_requests.filter(id.eq(project_request_id)))
        .set(project_request)
        .execute(conn)
}

pub fn delete(
    conn: &mut PgConnection,
    project_request_id: Uuid,
) -> Result<usize, diesel::result::Error> {
    diesel::delete(project_requests.filter(id.eq(project_request_id))).execute(conn)
}

#[cfg(test)]
mod tests {
    use camino::Utf8PathBuf;
    use chrono::{NaiveDate, NaiveDateTime, NaiveTime};
    use tempfile::tempdir;
    use uuid::Uuid;

    use crate::model::{
        operations::{
            project_requests::{delete, find_by_id, find_by_project, insert_with_id, update},
            tests::establish_connection,
        },
        structures::{
            project_requests::{
                ProjectRequest, ProjectRequestStatus, ProjectRequestType, UpdateProjectRequest,
            },
            projects::Project,
        },
    };

    #[test]
    fn test_model_insert_with_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let new_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(date, time),
            modification_date: None,
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: None,
            posix_user_id: None,
        };
        let inserted_project_request =
            insert_with_id(&mut connection, new_project_request.clone()).unwrap();
        assert_eq!(new_project_request, inserted_project_request);
    }

    #[test]
    fn test_model_find_by_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let new_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(date, time),
            modification_date: None,
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: None,
            posix_user_id: None,
        };
        let inserted_project_request =
            insert_with_id(&mut connection, new_project_request).unwrap();
        let project_request_found = find_by_id(&mut connection, inserted_project_request.id)
            .unwrap()
            .unwrap();
        assert_eq!(project_request_found, inserted_project_request);
    }

    #[test]
    fn test_find_by_project() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let temp_dir = tempdir().unwrap();
        let path: Utf8PathBuf = Utf8PathBuf::from_path_buf(temp_dir.path().to_owned()).unwrap();
        let project = Project {
            id: Uuid::new_v4(),
            name: String::from("New test project"),
            path,
            user_id: None,
        };

        let first_date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let first_time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let second_date = NaiveDate::from_ymd_opt(2021, 2, 2).unwrap();
        let second_time = NaiveTime::from_hms_milli_opt(1, 1, 1, 1).unwrap();
        let third_date = NaiveDate::from_ymd_opt(2022, 3, 3).unwrap();
        let third_time = NaiveTime::from_hms_milli_opt(2, 2, 2, 2).unwrap();

        // The first two extension dates are part of the project and the third one is not
        let first_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(first_date, first_time),
            modification_date: None,
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: Some(project.id),
            posix_user_id: None,
        };
        let second_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(second_date, second_time),
            modification_date: None,
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: Some(project.id),
            posix_user_id: None,
        };
        let third_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(third_date, third_time),
            modification_date: None,
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: None,
            posix_user_id: None,
        };
        crate::model::operations::projects::insert_with_id(&mut connection, project.clone())
            .unwrap();
        insert_with_id(&mut connection, first_project_request.clone()).unwrap();
        insert_with_id(&mut connection, second_project_request.clone()).unwrap();
        insert_with_id(&mut connection, third_project_request).unwrap();
        let found_project_requests = find_by_project(&mut connection, project.id).unwrap();
        assert_eq!(&found_project_requests.len(), &2);
        assert!(found_project_requests
            .iter()
            .any(|project_request| project_request.id == first_project_request.id));
        assert!(found_project_requests
            .iter()
            .any(|project_request| project_request.id == second_project_request.id));
    }

    #[test]
    fn test_find_by_project_invalid_project_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let projects_requests = find_by_project(&mut connection, Uuid::new_v4());
        assert!(projects_requests.is_err());
        assert_eq!(projects_requests, Err(diesel::result::Error::NotFound))
    }

    #[test]
    fn test_model_update() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let updated_date = NaiveDate::from_ymd_opt(2021, 2, 2).unwrap();
        let updated_time = NaiveTime::from_hms_milli_opt(1, 1, 1, 1).unwrap();
        let modification_date = NaiveDate::from_ymd_opt(2022, 3, 3).unwrap();
        let modification_time = NaiveTime::from_hms_milli_opt(2, 2, 2, 2).unwrap();
        let updated_modification_date = NaiveDate::from_ymd_opt(2023, 4, 4).unwrap();
        let updated_modification_time = NaiveTime::from_hms_milli_opt(3, 3, 3, 3).unwrap();
        let new_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(date, time),
            modification_date: Some(NaiveDateTime::new(modification_date, modification_time)),
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: None,
            posix_user_id: None,
        };
        let updated_project_request = UpdateProjectRequest {
            date: Some(NaiveDateTime::new(updated_date, updated_time)),
            modification_date: Some(NaiveDateTime::new(
                updated_modification_date,
                updated_modification_time,
            )),
            comment: Some(String::from("Updated comment")),
            project_request_type: Some(ProjectRequestType::Creation),
            project_request_status: Some(ProjectRequestStatus::Pending),
            project_id: None,
            posix_user_id: None,
        };
        let inserted_project_request =
            insert_with_id(&mut connection, new_project_request.clone()).unwrap();

        let updated_project_request_count = update(
            &mut connection,
            inserted_project_request.id,
            updated_project_request.clone(),
        )
        .unwrap();
        assert_eq!(updated_project_request_count, 1);

        let saved_project_request = find_by_id(&mut connection, inserted_project_request.id)
            .unwrap()
            .unwrap();

        assert_eq!(saved_project_request.id, new_project_request.id);
        assert_eq!(
            saved_project_request.date,
            updated_project_request.date.unwrap()
        );
        assert_eq!(
            saved_project_request.modification_date,
            updated_project_request.modification_date
        );
        assert_eq!(
            saved_project_request.comment,
            updated_project_request.comment
        );
        assert_eq!(
            saved_project_request.project_request_type,
            updated_project_request.project_request_type.unwrap()
        );
        assert_eq!(
            saved_project_request.project_id,
            updated_project_request.project_id
        );
    }

    #[test]
    fn test_model_delete() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let date = NaiveDate::from_ymd_opt(2020, 1, 1).unwrap();
        let time = NaiveTime::from_hms_milli_opt(0, 0, 0, 0).unwrap();
        let modification_date = NaiveDate::from_ymd_opt(2021, 2, 2).unwrap();
        let modification_time = NaiveTime::from_hms_milli_opt(1, 1, 1, 1).unwrap();
        let new_project_request = ProjectRequest {
            id: Uuid::new_v4(),
            date: NaiveDateTime::new(date, time),
            modification_date: Some(NaiveDateTime::new(modification_date, modification_time)),
            comment: Some(String::from("This is some comment")),
            project_request_type: ProjectRequestType::Quota,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: None,
            posix_user_id: None,
        };
        let inserted_project_request =
            insert_with_id(&mut connection, new_project_request).unwrap();
        find_by_id(&mut connection, inserted_project_request.id)
            .unwrap()
            .unwrap();

        let rows_affected = delete(&mut connection, inserted_project_request.id).unwrap();
        assert_eq!(rows_affected, 1);
        let project_request_found =
            find_by_id(&mut connection, inserted_project_request.id).unwrap();
        assert_eq!(project_request_found, None);
    }
}
