use crate::model::structures::projects::*;
use crate::schema::projects::dsl::projects;
use crate::schema::projects::id;
use diesel::prelude::*;
use uuid::Uuid;

pub fn insert_with_id(
    conn: &mut PgConnection,
    new_project: Project,
) -> Result<Project, diesel::result::Error> {
    use crate::schema::projects;
    let table_to_update = diesel::insert_into(projects::table);
    table_to_update
        .values(new_project)
        .returning(Project::as_returning())
        .get_result(conn)
}

pub fn find_by_id(conn: &mut PgConnection, project_id: Uuid) -> QueryResult<Option<Project>> {
    projects
        .filter(id.eq(project_id))
        .first::<Project>(conn)
        .optional()
}

pub fn find_by_user(conn: &mut PgConnection, user_id: i32) -> QueryResult<Vec<Project>> {
    let user = match crate::model::operations::users::find_by_id(conn, user_id) {
        Ok(found_user) => match found_user {
            Some(val) => val,
            None => return Err(diesel::result::Error::NotFound),
        },
        Err(err) => return Err(err),
    };
    let related_projects = match <Project as BelongingToDsl<
        &crate::model::structures::users::User,
    >>::belonging_to(&user)
    .load::<Project>(conn)
    {
        Ok(result) => result,
        Err(err) => return Err(err),
    };
    Ok(related_projects)
}

pub fn find_all(conn: &mut PgConnection) -> QueryResult<Vec<Project>> {
    projects.load::<Project>(conn)
}

pub fn update(
    conn: &mut PgConnection,
    project_id: uuid::Uuid,
    project: UpdateProject,
) -> Result<usize, diesel::result::Error> {
    diesel::update(projects.filter(id.eq(project_id)))
        .set(project)
        .execute(conn)
}

pub fn delete(conn: &mut PgConnection, project_id: Uuid) -> Result<usize, diesel::result::Error> {
    diesel::delete(projects.filter(id.eq(project_id))).execute(conn)
}

#[cfg(test)]
mod tests {
    use crate::model::{
        operations::{
            projects::{delete, find_by_id, insert_with_id, update},
            tests::establish_connection,
        },
        structures::{
            projects::{Project, UpdateProject},
            teams::Team,
            users::User,
        },
    };
    use camino::Utf8PathBuf;
    use uuid::Uuid;

    #[test]
    fn test_model_insert_with_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_team = Team {
            posix_id: 0,
            team_name: String::from("test_model_insert_with_id team_name"),
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        let new_user = User {
            posix_id: 1,
            username: String::from("test_model_insert_with_id username"),
            policy_accepted: true,
            team_id: team.posix_id,
        };
        let user =
            crate::model::operations::users::insert_with_id(&mut connection, new_user).unwrap();
        let id = Uuid::new_v4();
        let name = String::from("test_model_insert_with_id name");
        let path = Utf8PathBuf::from("/tmp/test");
        let uid = Some(user.posix_id);
        let new_project = Project {
            id,
            name: name.clone(),
            path: path.clone(),
            user_id: uid,
        };
        let inserted_project = insert_with_id(&mut connection, new_project).unwrap();
        assert_eq!(inserted_project.id, id);
        assert_eq!(inserted_project.name, name);
        assert_eq!(inserted_project.path, path);
        assert_eq!(inserted_project.user_id, uid);
    }

    #[test]
    fn test_model_find_by_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_team = Team {
            posix_id: 2,
            team_name: String::from("test_model_find_by_id team_name"),
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        let new_user = User {
            posix_id: 3,
            username: String::from("test_model_find_by_id username"),
            policy_accepted: true,
            team_id: team.posix_id,
        };
        let user =
            crate::model::operations::users::insert_with_id(&mut connection, new_user).unwrap();
        let id = Uuid::new_v4();
        let name = String::from("test_model_find_by_id name");
        let path = Utf8PathBuf::from("/tmp/test");
        let uid = Some(user.posix_id);
        let new_project = Project {
            id,
            name: name.clone(),
            path: path.clone(),
            user_id: uid,
        };
        let inserted_project = insert_with_id(&mut connection, new_project).unwrap();
        let project_found = find_by_id(&mut connection, inserted_project.id)
            .unwrap()
            .unwrap();
        assert_eq!(project_found.id, id);
        assert_eq!(project_found.name, name);
        assert_eq!(project_found.path, path);
        assert_eq!(project_found.user_id, uid);
    }

    #[test]
    fn test_model_update() {
        dotenvy::dotenv().ok();

        let mut connection = establish_connection();

        let team_name = String::from("test_model_update team_name");
        let team = crate::model::operations::teams::insert_with_id(
            &mut connection,
            Team {
                posix_id: 4,
                team_name,
            },
        )
        .unwrap();

        let posix_id = 5;
        let username = String::from("test_model_insert_with_id");
        let policy_accepted = false;
        let new_user = User {
            posix_id,
            username,
            policy_accepted,
            team_id: team.posix_id,
        };
        let inserted_user =
            crate::model::operations::users::insert_with_id(&mut connection, new_user).unwrap();

        let new_project = Project {
            id: Uuid::new_v4(),
            name: String::from("test_model_update new_name"),
            path: Utf8PathBuf::from("/tmp/test"),
            user_id: Some(inserted_user.posix_id),
        };

        let updated_name = String::from("test_model_update updated_name");
        let updated_path = Utf8PathBuf::from("/tmp/test/updated");
        let updated_project = UpdateProject {
            name: Some(updated_name.clone()),
            path: Some(updated_path.clone()),
            user_id: None,
        };

        let inserted_project = insert_with_id(&mut connection, new_project).unwrap();
        let updated_user_count =
            update(&mut connection, inserted_project.id, updated_project).unwrap();
        assert_eq!(updated_user_count, 1);

        let saved_project = find_by_id(&mut connection, inserted_project.id)
            .unwrap()
            .unwrap();
        assert_eq!(saved_project.id, inserted_project.id);
        assert_eq!(saved_project.name, updated_name);
        assert_eq!(saved_project.path, updated_path);
        assert_eq!(saved_project.user_id, Some(inserted_user.posix_id));
    }

    #[test]
    fn test_model_delete_user() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let new_team = Team {
            posix_id: 6,
            team_name: String::from("test_model_find_by_id team_name"),
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        let new_user = User {
            posix_id: 7,
            username: String::from("test_model_find_by_id username"),
            policy_accepted: true,
            team_id: team.posix_id,
        };
        let user =
            crate::model::operations::users::insert_with_id(&mut connection, new_user).unwrap();
        let id = Uuid::new_v4();
        let name = String::from("test_model_find_by_id name");
        let path = Utf8PathBuf::from("/tmp/test");
        let uid = Some(user.posix_id);
        let new_project = Project {
            id,
            name,
            path,
            user_id: uid,
        };
        let inserted_project = insert_with_id(&mut connection, new_project).unwrap();
        find_by_id(&mut connection, inserted_project.id)
            .unwrap()
            .unwrap();

        let rows_affected = delete(&mut connection, inserted_project.id).unwrap();
        assert_eq!(rows_affected, 1);
        let project_found = find_by_id(&mut connection, inserted_project.id).unwrap();
        assert_eq!(project_found, None);
    }
}
