use crate::model::structures::teams::*;
use crate::schema::teams::dsl::teams;
use crate::schema::teams::posix_id;
use diesel::prelude::*;

pub fn upsert(conn: &mut PgConnection, team: Team) -> Result<Team, diesel::result::Error> {
    use crate::schema::teams;
    let table_to_update = diesel::insert_into(teams::table);
    table_to_update
        .values(&team)
        .on_conflict(posix_id)
        .do_update()
        .set(&team)
        .returning(Team::as_returning())
        .get_result(conn)
}

pub fn insert_with_id(
    conn: &mut PgConnection,
    new_team: Team,
) -> Result<Team, diesel::result::Error> {
    use crate::schema::teams;
    let table_to_update = diesel::insert_into(teams::table);
    table_to_update
        .values(&new_team)
        .returning(Team::as_returning())
        .get_result(conn)
}

pub fn find_by_id(conn: &mut PgConnection, team_id: i32) -> QueryResult<Option<Team>> {
    teams
        .filter(posix_id.eq(team_id))
        .first::<Team>(conn)
        .optional()
}

pub fn delete(conn: &mut PgConnection, team_id: i32) -> Result<usize, diesel::result::Error> {
    diesel::delete(teams.filter(posix_id.eq(team_id))).execute(conn)
}

#[cfg(test)]
mod tests {
    use crate::model::{
        operations::{
            teams::{delete, find_by_id},
            tests::establish_connection,
        },
        structures::teams::Team,
    };

    #[test]
    fn test_model_insert_with_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let posix_id = 0;
        let team_name = String::from("test_model_insert_with_id");
        let new_team = Team {
            posix_id,
            team_name,
        };
        let inserted_team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        assert_eq!(inserted_team.posix_id, posix_id);
    }

    #[test]
    fn test_upsert() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let posix_id = 1;
        let team_name = String::from("test_upsert");
        let another_team_name = String::from("test_upsert another team name");
        let new_team = Team {
            posix_id,
            team_name: team_name.clone(),
        };
        let another_team = Team {
            posix_id,
            team_name: another_team_name.clone(),
        };
        let inserted_team =
            crate::model::operations::teams::upsert(&mut connection, new_team).unwrap();
        assert_eq!(inserted_team.posix_id, posix_id);
        assert_eq!(inserted_team.team_name, team_name);

        let updated_team =
            crate::model::operations::teams::upsert(&mut connection, another_team).unwrap();
        assert_eq!(updated_team.posix_id, posix_id);
        assert_eq!(updated_team.team_name, another_team_name);
        let found_team = find_by_id(&mut connection, posix_id).unwrap().unwrap();
        assert_eq!(found_team.posix_id, posix_id);
        assert_eq!(found_team.team_name, another_team_name);
    }

    #[test]
    fn test_model_find_by_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let posix_id = 2;
        let team_name = String::from("test_model_find_by_id");
        let new_team = Team {
            posix_id,
            team_name,
        };
        let inserted_team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        let team_found = find_by_id(&mut connection, inserted_team.posix_id)
            .unwrap()
            .unwrap();
        assert_eq!(team_found.posix_id, inserted_team.posix_id);
    }

    #[test]
    fn test_model_delete_user() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let posix_id = 3;

        let team_name = String::from("test_model_delete_user");
        let new_team = Team {
            posix_id,
            team_name,
        };
        let inserted_team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        find_by_id(&mut connection, inserted_team.posix_id)
            .unwrap()
            .unwrap();

        let rows_affected = delete(&mut connection, inserted_team.posix_id).unwrap();
        assert_eq!(rows_affected, 1);
        let team_found = find_by_id(&mut connection, inserted_team.posix_id).unwrap();
        assert_eq!(team_found, None);
    }
}
