use crate::model::structures::users::*;
use crate::schema::users::dsl::users;
use crate::schema::users::posix_id;
use diesel::prelude::*;

pub fn upsert(conn: &mut PgConnection, user: User) -> Result<User, diesel::result::Error> {
    use crate::schema::users;
    let table_to_update = diesel::insert_into(users::table);
    table_to_update
        .values(&user)
        .on_conflict(posix_id)
        .do_update()
        .set(&user)
        .returning(User::as_returning())
        .get_result(conn)
}

pub fn insert_with_id(
    conn: &mut PgConnection,
    new_user: User,
) -> Result<User, diesel::result::Error> {
    use crate::schema::users;
    let table_to_update = diesel::insert_into(users::table);
    table_to_update
        .values(&new_user)
        .returning(User::as_returning())
        .get_result(conn)
}

pub fn find_by_id(conn: &mut PgConnection, user_id: i32) -> QueryResult<Option<User>> {
    users
        .filter(posix_id.eq(user_id))
        .first::<User>(conn)
        .optional()
}

pub fn update(
    conn: &mut PgConnection,
    user_id: i32,
    user: UpdateUser,
) -> Result<usize, diesel::result::Error> {
    diesel::update(users.filter(posix_id.eq(user_id)))
        .set(user)
        .execute(conn)
}

pub fn delete(conn: &mut PgConnection, user_id: i32) -> Result<usize, diesel::result::Error> {
    diesel::delete(users.filter(posix_id.eq(user_id))).execute(conn)
}

#[cfg(test)]
mod tests {
    use crate::model::{
        operations::{
            tests::establish_connection,
            users::{delete, find_by_id, insert_with_id, update, upsert},
        },
        structures::{
            teams::Team,
            users::{UpdateUser, User},
        },
    };

    #[test]
    fn test_model_upsert() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let team_name = String::from("test_model_upsert team_name");
        let new_team = Team {
            posix_id: 0,
            team_name,
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        let posix_id = 1;
        let username = String::from("test_model_upsert username");
        let policy_accepted = false;
        let new_user = User {
            posix_id,
            username,
            policy_accepted,
            team_id: team.posix_id,
        };
        let inserted_user = upsert(&mut connection, new_user).unwrap();
        assert_eq!(inserted_user.posix_id, posix_id);
        assert_eq!(inserted_user.policy_accepted, policy_accepted);
        assert_eq!(inserted_user.team_id, team.posix_id);

        let updated_username = String::from("test_model_upsert updated username");
        let updated_policy_accepted = true;
        let updated_team = Team {
            posix_id: 2,
            team_name: String::from("test_model_upsert updated_team_name"),
        };

        let updated_team_insert =
            crate::model::operations::teams::insert_with_id(&mut connection, updated_team).unwrap();
        let updated_user = User {
            posix_id,
            username: updated_username.clone(),
            policy_accepted: updated_policy_accepted,
            team_id: updated_team_insert.posix_id,
        };
        crate::model::operations::users::upsert(&mut connection, updated_user).unwrap();
        let found_user = find_by_id(&mut connection, posix_id).unwrap().unwrap();
        assert_eq!(found_user.posix_id, posix_id);
        assert_eq!(found_user.policy_accepted, updated_policy_accepted);
        assert_eq!(found_user.team_id, updated_team_insert.posix_id);
        assert_eq!(found_user.username, updated_username);
    }

    #[test]
    fn test_model_insert_with_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();
        let team_name = String::from("test_model_insert_with_id team_name");
        let new_team = Team {
            posix_id: 3,
            team_name,
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();
        let posix_id = 4;
        let username = String::from("test_model_insert_with_id username");
        let policy_accepted = false;
        let new_user = User {
            posix_id,
            username,
            policy_accepted,
            team_id: team.posix_id,
        };
        let inserted_user = insert_with_id(&mut connection, new_user).unwrap();
        assert_eq!(inserted_user.posix_id, posix_id);
        assert_eq!(inserted_user.policy_accepted, policy_accepted);
        assert_eq!(inserted_user.team_id, team.posix_id);
    }

    #[test]
    fn test_model_find_by_id() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();

        let team_name = String::from("test_model_find_by_id");
        let new_team = Team {
            posix_id: 5,
            team_name,
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();

        let posix_id = 6;
        let policy_accepted = false;
        let username = String::from("test_model_insert_with_id username");
        let new_user = User {
            posix_id,
            username: username.clone(),
            policy_accepted,
            team_id: team.posix_id,
        };
        let inserted_user = insert_with_id(&mut connection, new_user).unwrap();
        let user_found = find_by_id(&mut connection, inserted_user.posix_id)
            .unwrap()
            .unwrap();
        assert_eq!(user_found.posix_id, posix_id);
        assert_eq!(user_found.username, username);
        assert_eq!(user_found.policy_accepted, policy_accepted);
        assert_eq!(user_found.team_id, team.posix_id);
    }

    #[test]
    fn test_model_update() {
        dotenvy::dotenv().ok();

        let mut connection = establish_connection();

        let team_name = String::from("test_model_update team_name");
        let team = crate::model::operations::teams::insert_with_id(
            &mut connection,
            Team {
                posix_id: 7,
                team_name,
            },
        )
        .unwrap();

        let other_team_name = String::from("test_model_update other_team_name");
        let other_team = crate::model::operations::teams::insert_with_id(
            &mut connection,
            Team {
                posix_id: 8,
                team_name: other_team_name,
            },
        )
        .unwrap();

        let posix_id = 7;
        let username = String::from("test_model_insert_with_id");
        let policy_accepted = false;
        let new_user = User {
            posix_id,
            username,
            policy_accepted,
            team_id: team.posix_id,
        };

        let new_username = String::from("test_model_update_new");
        let updated_policy_accepted = true;
        let updated_user = UpdateUser {
            username: Some(new_username.clone()),
            team_id: Some(other_team.posix_id),
            policy_accepted: Some(updated_policy_accepted),
        };
        let inserted_user = insert_with_id(&mut connection, new_user).unwrap();
        let updated_user_count =
            update(&mut connection, inserted_user.posix_id, updated_user).unwrap();
        assert_eq!(updated_user_count, 1);

        let saved_user = find_by_id(&mut connection, inserted_user.posix_id)
            .unwrap()
            .unwrap();
        assert_eq!(saved_user.posix_id, inserted_user.posix_id);
        assert_eq!(saved_user.username, new_username);
        assert_eq!(saved_user.policy_accepted, updated_policy_accepted);
        assert_eq!(saved_user.team_id, other_team.posix_id);
    }

    #[test]
    fn test_model_delete_user() {
        dotenvy::dotenv().ok();
        let mut connection = establish_connection();

        let team_name = String::from("test_model_insert_with_id team_name");
        let new_team = Team {
            posix_id: 9,
            team_name,
        };
        let team =
            crate::model::operations::teams::insert_with_id(&mut connection, new_team).unwrap();

        let posix_id = 10;
        let username = String::from("test_model_insert_with_id username");
        let policy_accepted = false;
        let new_user = User {
            posix_id,
            username,
            policy_accepted,
            team_id: team.posix_id,
        };
        let inserted_user = insert_with_id(&mut connection, new_user).unwrap();
        find_by_id(&mut connection, inserted_user.posix_id)
            .unwrap()
            .unwrap();

        let rows_affected = delete(&mut connection, inserted_user.posix_id).unwrap();
        assert_eq!(rows_affected, 1);
        let user_found = find_by_id(&mut connection, inserted_user.posix_id).unwrap();
        assert_eq!(user_found, None);
    }
}
