use chrono::{Local, NaiveDateTime};
use serde::{Deserialize, Serialize};
use uuid::Uuid;

use crate::model::structures::projects::Project;
use crate::schema::project_requests::{self};

#[derive(diesel_derive_enum::DbEnum, Debug, Serialize, Deserialize, Eq, PartialEq, Clone, Copy)]
#[ExistingTypePath = "crate::schema::sql_types::ProjectRequestTypes"]
pub enum ProjectRequestType {
    Quota,
    Duration,
    Creation,
    Reservation,
}

#[derive(diesel_derive_enum::DbEnum, Debug, Serialize, Deserialize, Eq, PartialEq, Clone, Copy)]
#[ExistingTypePath = "crate::schema::sql_types::ProjectRequestStatus"]
pub enum ProjectRequestStatus {
    Validated,
    Refused,
    Pending,
    Canceled,
}

#[derive(
    Selectable,
    Queryable,
    Identifiable,
    Associations,
    Insertable,
    Deserialize,
    Serialize,
    Eq,
    PartialEq,
    Debug,
    Clone,
)]
#[diesel(table_name = project_requests)]
#[diesel(primary_key(id))]
#[diesel(belongs_to(Project))]
pub struct ProjectRequest {
    pub id: Uuid,
    pub date: NaiveDateTime,
    pub modification_date: Option<NaiveDateTime>,
    pub comment: Option<String>,
    pub project_request_type: ProjectRequestType,
    pub project_request_status: ProjectRequestStatus,
    pub project_id: Option<Uuid>,
    pub posix_user_id: Option<i32>,
}

#[derive(Insertable, Deserialize)]
#[diesel(table_name = project_requests)]
pub struct NewProjectRequest {
    pub date: NaiveDateTime,
    pub modification_date: Option<NaiveDateTime>,
    pub comment: Option<String>,
    pub project_request_type: ProjectRequestType,
    pub project_request_status: ProjectRequestStatus,
    pub project_id: Option<Uuid>,
    pub posix_user_id: Option<i32>,
}

#[derive(AsChangeset, Deserialize, Clone)]
#[diesel(table_name = project_requests)]
pub struct UpdateProjectRequest {
    pub date: Option<NaiveDateTime>,
    pub modification_date: Option<NaiveDateTime>,
    pub comment: Option<String>,
    pub project_request_type: Option<ProjectRequestType>,
    pub project_request_status: Option<ProjectRequestStatus>,
    pub project_id: Option<Uuid>,
    pub posix_user_id: Option<i32>,
}

impl NewProjectRequest {
    /// Creates a new project request of a type [`ProjectRequestType::Creation`] with date and modification_date set to now and the project_id to `None` since the project has not been created yet, it is created once the request has been validated
    ///
    /// *`comment` - An optional comment added to the project creation request
    ///
    /// *`posix_user_id` - The id of the user making the request
    pub fn new_project_request_type_creation(
        comment: Option<String>,
        posix_user_id: Option<i32>,
    ) -> Self {
        let now = Local::now().naive_local();
        Self {
            date: now,
            modification_date: Some(now),
            comment,
            project_request_type: ProjectRequestType::Creation,
            project_request_status: ProjectRequestStatus::Pending,
            project_id: None,
            posix_user_id,
        }
    }
}
