use actix_web::{delete, error, get, post, put, web, HttpResponse, Responder};
use dmp_dto::project_requests::NewProjectRequestDTO;
use uuid::Uuid;

use crate::{
    connections::ConnectionsHandle,
    model::structures::project_requests::{
        NewProjectRequest, ProjectRequest, UpdateProjectRequest,
    },
};

#[post("/{project_request_id}")]
async fn insert_with_id(
    _: crate::router::auth::JWTClaims,
    connections: ConnectionsHandle,
    project_request_id: web::Path<Uuid>,
    pr: web::Json<NewProjectRequest>,
) -> actix_web::Result<impl Responder> {
    let request = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        let new_project_request = ProjectRequest {
            id: project_request_id.into_inner(),
            date: pr.date,
            modification_date: pr.modification_date,
            comment: pr.comment.clone(),
            project_request_type: pr.project_request_type,
            project_request_status: pr.project_request_status,
            project_id: pr.project_id,
            posix_user_id: pr.posix_user_id,
        };
        crate::model::operations::project_requests::insert_with_id(
            &mut connection,
            new_project_request,
        )
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    Ok(HttpResponse::Ok().json(request))
}

#[post("/project_requests")]
async fn insert(
    claims: crate::router::auth::JWTClaims,
    connections: ConnectionsHandle,
    pr: web::Json<NewProjectRequestDTO>,
) -> actix_web::Result<impl Responder> {
    let request = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        let pr = pr.into_inner();
        let project_request =
            NewProjectRequest::new_project_request_type_creation(pr.comment, Some(claims.user_id));
        crate::model::operations::project_requests::insert(&mut connection, project_request)
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    Ok(HttpResponse::Ok().json(request))
}

#[get("/{request_id}")]
async fn find_by_id(
    connections: ConnectionsHandle,
    request_id: web::Path<Uuid>,
) -> actix_web::Result<impl Responder> {
    let id = request_id.into_inner();
    let request = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        crate::model::operations::project_requests::find_by_id(&mut connection, id)
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    match request {
        Some(request) => Ok(HttpResponse::Ok().json(request)),
        None => Ok(HttpResponse::NotFound().into()),
    }
}

#[get("/projects/{project_id}")]
async fn find_by_project(
    connections: ConnectionsHandle,
    project_id: web::Path<Uuid>,
) -> actix_web::Result<impl Responder> {
    let p_id = project_id.into_inner();
    let request = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        crate::model::operations::project_requests::find_by_project(&mut connection, p_id)
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    Ok(HttpResponse::Ok().json(request))
}

#[put("/{request_id}")]
async fn update(
    connections: ConnectionsHandle,
    user_id: web::Path<Uuid>,
    updated_request_info: web::Json<UpdateProjectRequest>,
) -> actix_web::Result<impl Responder> {
    let request_updated = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Coulnd't get db connection from pool");
        crate::model::operations::project_requests::update(
            &mut connection,
            user_id.into_inner(),
            updated_request_info.into_inner(),
        )
    })
    .await?
    .map_err(error::ErrorInternalServerError);
    match request_updated {
        Ok(requests_updated_count) => Ok(HttpResponse::Ok().json(requests_updated_count)),
        Err(error) => Err(error),
    }
}

#[delete("/{request_id}")]
async fn delete(
    connections: ConnectionsHandle,
    request_id: web::Path<Uuid>,
) -> actix_web::Result<impl Responder> {
    let id = request_id.into_inner();
    let deleted_request_count = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        crate::model::operations::project_requests::delete(&mut connection, id)
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    match deleted_request_count {
        0 => Ok(HttpResponse::NotFound()),
        _ => Ok(HttpResponse::Ok()),
    }
}
