use actix_web::{delete, error, get, post, put, web, HttpResponse, Responder};

use crate::{
    connections::ConnectionsHandle,
    model::structures::{
        teams::Team,
        users::{UpdateUser, User},
    },
    router::{auth::JWTClaims, models::UserWithTeam},
};

#[post("/users/add_user_and_team")]
async fn add_user_and_team(
    _: JWTClaims,
    connections: ConnectionsHandle,
    user_with_team: web::Json<UserWithTeam>,
) -> actix_web::Result<impl Responder> {
    let team: Team = user_with_team.clone().into();
    let connections = connections.lock().unwrap();
    let mut connection = connections
        .db
        .get()
        .expect("Couldn't get db connection from pool");
    let upsert_team = crate::model::operations::teams::upsert(&mut connection, team);
    match upsert_team {
        Ok(team) => {
            let user: User = user_with_team.into_inner().into();
            match crate::model::operations::users::upsert(&mut connection, user) {
                Ok(user) => {
                    let result = UserWithTeam {
                        posix_team_id: team.posix_id,
                        team_name: team.team_name,
                        posix_user_id: user.posix_id,
                        username: user.username,
                        policy_accepted: user.policy_accepted,
                    };
                    Ok(HttpResponse::Ok().json(result))
                }
                Err(error) => Ok(HttpResponse::InternalServerError().json(error.to_string())),
            }
        }
        Err(error) => Ok(HttpResponse::InternalServerError().json(error.to_string())),
    }
}

#[post("")]
async fn insert_with_id(
    _: crate::router::auth::JWTClaims,
    connections: ConnectionsHandle,
    u: web::Json<User>,
) -> actix_web::Result<impl Responder> {
    let user = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        crate::model::operations::users::insert_with_id(&mut connection, u.into_inner())
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    Ok(HttpResponse::Ok().json(user))
}

#[get("/{user_id}")]
async fn find_by_id(
    connections: ConnectionsHandle,
    user_id: web::Path<i32>,
) -> actix_web::Result<impl Responder> {
    let id = user_id.into_inner();
    let user = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        crate::model::operations::users::find_by_id(&mut connection, id)
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    match user {
        Some(user) => Ok(HttpResponse::Ok().json(user)),
        None => Ok(HttpResponse::NotFound().into()),
    }
}

#[put("/{user_id}")]
async fn update(
    connections: ConnectionsHandle,
    user_id: web::Path<i32>,
    updated_user_info: web::Json<UpdateUser>,
) -> actix_web::Result<impl Responder> {
    let users_updated = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Coulnd't get db connection from pool");
        crate::model::operations::users::update(
            &mut connection,
            user_id.into_inner(),
            updated_user_info.into_inner(),
        )
    })
    .await?
    .map_err(error::ErrorInternalServerError);
    match users_updated {
        Ok(users_updated_count) => Ok(HttpResponse::Ok().json(users_updated_count)),
        Err(error) => Err(error),
    }
}

#[delete("/{user_id}")]
async fn delete(
    connections: ConnectionsHandle,
    user_id: web::Path<i32>,
) -> actix_web::Result<impl Responder> {
    let id = user_id.into_inner();
    let deleted_user_count = web::block(move || {
        let connections = connections.lock().unwrap();
        let mut connection = connections
            .db
            .get()
            .expect("Couldn't get db connection from pool");
        crate::model::operations::users::delete(&mut connection, id)
    })
    .await?
    .map_err(error::ErrorInternalServerError)?;
    match deleted_user_count {
        0 => Ok(HttpResponse::NotFound()),
        _ => Ok(HttpResponse::Ok()),
    }
}
