use openidconnect::{
    core::{
        CoreAuthDisplay, CoreAuthPrompt, CoreClient, CoreErrorResponseType, CoreGenderClaim,
        CoreIdToken, CoreJsonWebKey, CoreJsonWebKeyType, CoreJsonWebKeyUse,
        CoreJweContentEncryptionAlgorithm, CoreJwsSigningAlgorithm, CoreResponseType,
        CoreRevocableToken, CoreRevocationErrorResponse, CoreTokenIntrospectionResponse,
        CoreTokenType,
    },
    reqwest::async_http_client,
    url::Url,
    AccessToken, AccessTokenHash, AdditionalClaims, AuthenticationFlow, AuthorizationCode,
    ClaimsVerificationError, Client, ClientId, CsrfToken, EmptyExtraTokenFields, IdToken,
    IdTokenFields, IssuerUrl, LogoutRequest, Nonce, ProviderMetadataWithLogout, RedirectUrl,
    RefreshToken, RequestTokenError, StandardErrorResponse, StandardTokenResponse,
};
use serde::{Deserialize, Serialize};

use crate::{errors::Error, props::client::ClientProps, DIOXUS_FRONT_CLIENT_ID};

pub type DMPIdToken = IdToken<
    DMPClaims,
    CoreGenderClaim,
    CoreJweContentEncryptionAlgorithm,
    CoreJwsSigningAlgorithm,
    CoreJsonWebKeyType,
>;

pub type DMPIdTokenFields = IdTokenFields<
    DMPClaims,
    EmptyExtraTokenFields,
    CoreGenderClaim,
    CoreJweContentEncryptionAlgorithm,
    CoreJwsSigningAlgorithm,
    CoreJsonWebKeyType,
>;

pub type DMPTokenResponse = StandardTokenResponse<DMPIdTokenFields, CoreTokenType>;

pub type DMPClient = openidconnect::Client<
    DMPClaims,
    CoreAuthDisplay,
    CoreGenderClaim,
    CoreJweContentEncryptionAlgorithm,
    CoreJwsSigningAlgorithm,
    CoreJsonWebKeyType,
    CoreJsonWebKeyUse,
    CoreJsonWebKey,
    CoreAuthPrompt,
    StandardErrorResponse<CoreErrorResponseType>,
    DMPTokenResponse,
    CoreTokenType,
    CoreTokenIntrospectionResponse,
    CoreRevocableToken,
    CoreRevocationErrorResponse,
>;

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct DMPClaims {
    // /// The posix id of the user, in our case this is the ldap id
    #[serde(alias = "uid_number")]
    user_id: i32,
    /// The posix id of the group the user belongs to
    #[serde(alias = "gid_number")]
    team_id: i32,
}

impl AdditionalClaims for DMPClaims {}

#[derive(Clone, Debug, Default)]
pub struct ClientState {
    pub oidc_client: Option<ClientProps>,
}

/// State that holds the nonce and authorization url and the nonce generated to log in an user
#[derive(Clone, Deserialize, Serialize, Default)]
pub struct AuthRequestState {
    pub auth_request: Option<AuthRequest>,
}

#[derive(Clone, Deserialize, Serialize)]
pub struct AuthRequest {
    pub nonce: Nonce,
    pub authorize_url: String,
}

/// State the tokens returned once the user is authenticated
#[derive(Debug, Deserialize, Serialize, Default, Clone)]
pub struct AuthTokenState {
    /// Token used to identify the user
    pub id_token: Option<DMPIdToken>,
    /// Token used to refresh the tokens if they expire
    pub refresh_token: Option<RefreshToken>,
    /// Token used to access ressources
    pub access_token: Option<AccessToken>,
}

pub fn email(
    client: DMPClient,
    id_token: DMPIdToken,
    nonce: Nonce,
) -> Result<String, ClaimsVerificationError> {
    match id_token.claims(&client.id_token_verifier(), &nonce) {
        Ok(claims) => Ok(claims.clone().email().unwrap().to_string()),
        Err(error) => Err(error),
    }
}

/// Returns the POSIX user id returned from the claims
pub fn user_id(
    client: DMPClient,
    id_token: DMPIdToken,
    nonce: Nonce,
) -> Result<i32, ClaimsVerificationError> {
    let claims = id_token
        .claims(&client.id_token_verifier(), &nonce)?
        .additional_claims()
        .user_id;
    Ok(claims)
}

/// Returns the POSIX team id returned from the claims
pub fn team_id(
    client: DMPClient,
    id_token: DMPIdToken,
    nonce: Nonce,
) -> Result<i32, ClaimsVerificationError> {
    let claims = id_token
        .claims(&client.id_token_verifier(), &nonce)?
        .additional_claims()
        .team_id;
    Ok(claims)
}

/// Generates the authorize url used to log in the user
/// To get a list of the available scopes, go to: /realms/{realm_name}/.well-known/openid-configuration
pub fn authorize_url(client: DMPClient) -> AuthRequest {
    let (authorize_url, _csrf_state, nonce) = client
        .authorize_url(
            AuthenticationFlow::<CoreResponseType>::AuthorizationCode,
            CsrfToken::new_random,
            Nonce::new_random,
        )
        .add_scope(openidconnect::Scope::new("email".to_string()))
        .add_scope(openidconnect::Scope::new("profile".to_string()))
        .add_scope(openidconnect::Scope::new("posix".to_string()))
        .url();
    AuthRequest {
        authorize_url: authorize_url.to_string(),
        nonce,
    }
}

pub async fn init_provider_metadata() -> Result<ProviderMetadataWithLogout, crate::errors::Error> {
    let issuer_url = IssuerUrl::new(crate::DIOXUS_FRONT_ISSUER_URL.to_string())?;
    Ok(ProviderMetadataWithLogout::discover_async(issuer_url, async_http_client).await?)
}

pub async fn init_oidc_client() -> Result<(ClientId, DMPClient), crate::errors::Error> {
    let client_id = ClientId::new(crate::DIOXUS_FRONT_CLIENT_ID.to_string());
    let provider_metadata = init_provider_metadata().await?;
    let client_secret = None;
    let redirect_url = RedirectUrl::new(format!("{}/login", crate::DIOXUS_FRONT_URL))?;

    Ok((
        client_id.clone(),
        Client::from_provider_metadata(provider_metadata, client_id, client_secret)
            .set_redirect_uri(redirect_url),
    ))
}

///TODO: Add pkce_pacifier
pub async fn token_response(
    oidc_client: DMPClient,
    code: String,
) -> Result<DMPTokenResponse, crate::errors::Error> {
    // let (pkce_challenge, pkce_verifier) = PkceCodeChallenge::new_random_sha256();
    Ok(oidc_client
        .exchange_code(AuthorizationCode::new(code.clone()))
        // .set_pkce_verifier(pkce_verifier)
        .request_async(async_http_client)
        .await?)
}

pub async fn exchange_refresh_token(
    oidc_client: DMPClient,
    refresh_token: RefreshToken,
) -> Result<
    DMPTokenResponse,
    RequestTokenError<
        openidconnect::reqwest::Error<reqwest::Error>,
        StandardErrorResponse<CoreErrorResponseType>,
    >,
> {
    oidc_client
        .exchange_refresh_token(&refresh_token)
        .request_async(async_http_client)
        .await
}

pub async fn log_out_url(id_token_hint: DMPIdToken) -> Result<Url, crate::errors::Error> {
    let provider_metadata = init_provider_metadata().await?;
    let end_session_url = provider_metadata
        .additional_metadata()
        .clone()
        .end_session_endpoint
        .unwrap();
    let logout_request: LogoutRequest = LogoutRequest::from(end_session_url);
    Ok(logout_request
        .set_client_id(ClientId::new(DIOXUS_FRONT_CLIENT_ID.to_string()))
        .set_id_token_hint(&id_token_hint)
        .http_get_url())
}

#[allow(dead_code)]
pub fn verify_access_token(
    client: CoreClient,
    id_token: CoreIdToken,
    nonce: Nonce,
    access_token: AccessToken,
) -> Result<(), Error> {
    let claims = id_token.claims(&client.id_token_verifier(), &nonce)?;
    if let Some(expected_access_token_hash) = claims.access_token_hash() {
        let actual_access_token_hash =
            AccessTokenHash::from_token(&access_token, &id_token.signing_alg()?)?;
        if actual_access_token_hash != *expected_access_token_hash {
            return Err(Error::AccessTokenMismatch {
                expected: expected_access_token_hash.clone(),
                found: actual_access_token_hash,
            });
        }
    }
    Ok(())
}
